# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['charmonium', 'charmonium.freeze']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'charmonium.freeze',
    'version': '0.5.5',
    'description': 'Injectively, deterministically maps arbitrary objects to hashable values',
    'long_description': '==========================\ncharmonium.freeze\n==========================\n\n.. image:: https://img.shields.io/pypi/v/charmonium.freeze\n   :alt: PyPI Package\n   :target: https://pypi.org/project/charmonium.freeze\n.. image:: https://img.shields.io/pypi/dm/charmonium.freeze\n   :alt: PyPI Downloads\n   :target: https://pypi.org/project/charmonium.freeze\n.. image:: https://img.shields.io/pypi/l/charmonium.freeze\n   :alt: License\n   :target: https://github.com/charmoniumQ/charmonium.freeze/blob/main/LICENSE\n.. image:: https://img.shields.io/pypi/pyversions/charmonium.freeze\n   :alt: Python Versions\n   :target: https://pypi.org/project/charmonium.freeze\n.. image:: https://img.shields.io/librariesio/sourcerank/pypi/charmonium.freeze\n   :alt: libraries.io sourcerank\n   :target: https://libraries.io/pypi/charmonium.freeze\n.. image:: https://img.shields.io/github/stars/charmoniumQ/charmonium.freeze?style=social\n   :alt: GitHub stars\n   :target: https://github.com/charmoniumQ/charmonium.freeze\n.. image:: https://github.com/charmoniumQ/charmonium.freeze/actions/workflows/main.yaml/badge.svg\n   :alt: CI status\n   :target: https://github.com/charmoniumQ/charmonium.freeze/actions/workflows/main.yaml\n.. image:: https://img.shields.io/github/last-commit/charmoniumQ/charmonium.determ_hash\n   :alt: GitHub last commit\n   :target: https://github.com/charmoniumQ/charmonium.freeze/commits\n.. image:: http://www.mypy-lang.org/static/mypy_badge.svg\n   :target: https://mypy.readthedocs.io/en/stable/\n   :alt: Checked with Mypy\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Code style: black\n\nInjectively, deterministically maps arbitrary objects to hashable, immutable values\n\n\n----------\nQuickstart\n----------\n\nIf you don\'t have ``pip`` installed, see the `pip install guide`_.\n\n.. _`pip install guide`: https://pip.pypa.io/en/latest/installing/\n\n.. code-block:: console\n\n    $ pip install charmonium.freeze\n\nFor a related project, |charmonium.cache|_, I needed a function that\ndeterministically, injectively maps objects to hashable objects.\n\n- "Injectively" means ``freeze(a) == freeze(b)`` implies ``a == b``\n  (with the precondition that ``a`` and ``b`` are of the same type).\n\n- "Deterministically" means it should return the same value **across\n  subsequent process invocations** (with the same interpreter major\n  and minor version), unlike Python\'s |hash|_ function, which is not\n  deterministic between processes.\n\n- "Hashable" means one can call ``hash(...)`` on it. All hashable\n  values are immutable.\n\n.. |hash| replace:: ``hash``\n.. _`hash`: https://docs.python.org/3.8/reference/datamodel.html#object.__hash__\n.. |charmonium.cache| replace:: ``charmonium.cache``\n.. _`charmonium.cache`: https://github.com/charmoniumQ/charmonium.cache\n\nHave you ever felt like you wanted to "freeze" a list of arbitrary\ndata into a hashable value? Now you can.\n\n>>> obj = [1, 2, 3, {4, 5, 6}, object()]\n>>> hash(obj)\nTraceback (most recent call last):\n  File "<stdin>", line 1, in <module>\nTypeError: unhashable type: \'list\'\n\n>>> from charmonium.freeze import freeze\n>>> from pprint import pprint\n>>> freeze(obj)\n(1, 2, 3, frozenset({4, 5, 6}), ((\'args\', (\'object\',)),))\n\nIf you want to actually boil this down into a single integer, see\n|charmonium.determ_hash|_. This library\'s job is just to freeze the\nstate.\n\n.. |charmonium.determ_hash| replace:: ``charmonium.determ_hash``\n.. _`charmonium.determ_hash`: https://github.com/charmoniumQ/charmonium.determ_hash\n\nIt even works on custom types.\n\n>>> # Make a custom type\n>>> class Struct:\n...     def frobnicate(self):\n...         print(123)\n>>> s = Struct()\n>>> s.attr = 4\n>>> freeze(s)\n((\'args\', (\'Struct\',)), (\'state\', ((\'attr\', 4),)))\n\nAnd methods, functions, lambdas, etc.\n\n>>> pprint(freeze(lambda x: x + 123))\n((\'code\',\n  ((\'name\', \'<lambda>\'),\n   (\'varnames\', (\'x\',)),\n   (\'constants\', (None, 123)),\n   (\'bytecode\', b\'|\\x00d\\x01\\x17\\x00S\\x00\'))),)\n>>> import functools\n>>> pprint(freeze(functools.partial(print, 123)))\n((\'constructor\', \'partial\'),\n (\'args\', (\'print\',)),\n (\'state\', (\'print\', (123,), (), None)))\n>>> pprint(freeze(Struct.frobnicate))\n((\'code\',\n  ((\'name\', \'frobnicate\'),\n   (\'varnames\', (\'self\',)),\n   (\'constants\', (None, 123)),\n   (\'bytecode\', b\'t\\x00d\\x01\\x83\\x01\\x01\\x00d\\x00S\\x00\'))),)\n>>> i = 0\n>>> def square_plus_i(x):\n...     # Value of global variable will be included in the function\'s frozen state.\n...     return x**2 + i\n... \n>>> pprint(freeze(square_plus_i))\n((\'code\',\n  ((\'name\', \'square_plus_i\'),\n   (\'varnames\', (\'x\',)),\n   (\'constants\', (None, 2)),\n   (\'bytecode\', b\'|\\x00d\\x01\\x13\\x00t\\x00\\x17\\x00S\\x00\'))),\n (\'closure globals\', ((\'i\', 0),)))\n\nIf the source code of ``square_plus_i`` changes between successive invocations,\nthen the ``freeze`` value will change. This is useful for caching unchanged\nfunctions.\n\n-------------\nSpecial cases\n-------------\n\n- ``freeze`` on functions returns their bytecode, constants, and\n  closure-vars. The remarkable thing is that this is true across subsequent\n  invocations of the same process. If the user edits the script and changes the\n  function, then it\'s ``freeze`` will change too.\n\n  ::\n\n    (freeze(f) == freeze(g)) implies (for all x, f(x) == g(x))\n\n- ``freeze`` on an object returns the data that used in the `pickle\n  protocol`_. This makes ``freeze`` work correctly on most user-defined\n  types. However, there can still be special cases: ``pickle`` may incorporate\n  non-deterministic values. In this case, there are two remedies:\n\n  - If you can tweak the definition of the class, add a method called\n    ``__getfrozenstate__`` which returns a deterministic snapshot of the\n    state. This takes precedence over the Pickle protocol, if it is defined.\n\n    >>> class Struct:\n    ...     pass\n    >>> s = Struct()\n    >>> s.attr = 4\n    >>> pprint(freeze(s))\n    ((\'args\', (\'Struct\',)), (\'state\', ((\'attr\', 4),)))\n    >>> # which is based on the Pickle protocol\'s definition of `__reduce__`:\n    >>> pprint(s.__reduce__())\n    (<function _reconstructor at 0x...>,\n     (<class \'__main__.Struct\'>, <class \'object\'>, None),\n     {\'attr\': 4})\n\n\n  - If you cannot tweak the definition of the class, you can still register `single dispatch handler`_ \n    for that type:\n\n    >>> from typing import Set, Hashable\n    >>> from charmonium.freeze import freeze, _freeze_dispatch, _freeze\n    >>> class Test:\n    ...     deterministic_val = 3\n    ...     nondeterministic_val = 4\n    ... \n    >>> @_freeze_dispatch.register(Test)\n    ... def _(obj: Test, tabu: Set[int], level: int) -> Hashable:\n    ...     # Type annotations are optional.\n    ...     # I have included them here for clarity.\n    ... \n    ...     # `tabu` is for object cycle detection.\n    ...     tabu = tabu | {id(obj)}\n    ... \n    ...     # `level` is for logging and infinite recursion detection.\n    ...     level = level + 1\n    ... \n    ...     # Freeze should depend only on deterministic values.\n    ...     if isinstance(obj.deterministic_val, int):\n    ...         return obj.deterministic_val\n    ...     else:\n    ...         # If the underlying instance variable is not hashable, we can use recursion to help.\n    ...         # Call `_freeze` instead of `freeze` to recurse with `tabu` and `level`.\n    ...         return _freeze(obj.deterministic_val, tabu, level)\n    ... \n    >>> freeze(Test())\n    3\n\n- Note that as of Python 3.7, dictionaries "remember" their insertion order. As such,\n\n  >>> freeze({"a": 1, "b": 2})\n  ((\'a\', 1), (\'b\', 2))\n  >>> freeze({"b": 2, "a": 1})\n  ((\'b\', 2), (\'a\', 1))\n\n.. _`pickle protocol`: https://docs.python.org/3/library/pickle.html#pickling-class-instances\n.. _`single dispatch handler`: https://docs.python.org/3/library/functools.html#functools.singledispatch\n\n----------\nDeveloping\n----------\n\nSee `CONTRIBUTING.md`_ for instructions on setting up a development environment.\n\n.. _`CONTRIBUTING.md`: https://github.com/charmoniumQ/charmonium.freeze/tree/main/CONTRIBUTING.md\n\n---------\nDebugging\n---------\n\nUse the following lines to see how ``freeze`` decomposes an object into\nprimitive values.\n\n.. code:: python\n\n    import logging, os\n    logger = logging.getLogger("charmonium.freeze")\n    logger.setLevel(logging.DEBUG)\n    fh = logging.FileHandler("freeze.log")\n    fh.setLevel(logging.DEBUG)\n    fh.setFormatter(logging.Formatter("%(message)s"))\n    logger.addHandler(fh)\n    logger.debug("Program %d", os.getpid())\n\n    i = 0\n    def square_plus_i(x):\n        # Value of global variable will be included in the function\'s frozen state.\n        return x**2 + i\n\n    from charmonium.freeze import freeze\n    freeze(square_plus_i)\n\n\nThis produces a log such as in ``freeze.log``:\n\n::\n\n    freeze begin <function square_plus_i at 0x7f9228bff550>\n     function <function square_plus_i at 0x7f9228bff550>\n      tuple ((\'code\', <code object square_plus_i at 0x7f9228c6cf50, file "/tmp/ipython_edit_303agyiz/ipython_edit_rez33yf_.py", line 2>), \'closure globals\', {\'i\': 0})\n       tuple (\'code\', <code object square_plus_i at 0x7f9228c6cf50, file "/tmp/ipython_edit_303agyiz/ipython_edit_rez33yf_.py", line 2>)\n        \'code\'\n        code <code object square_plus_i at 0x7f9228c6cf50, file "/tmp/ipython_edit_303agyiz/ipython_edit_rez33yf_.py", line 2>\n         tuple (None, 2)\n          None\n          2\n         b\'|\\x00d\\x01\\x13\\x00t\\x00\\x17\\x00S\\x00\'\n       \'closure globals\'\n       dict {\'i\': 0}\n        \'i\'\n        0\n    freeze end\n\nI do this to find the differences between subsequent runs:\n\n.. code:: shell\n\n    $ python code.py\n    $ mv freeze.log freeze.0.log\n\n    $ python code.py\n    $ mv freeze.log freeze.1.log\n\n    $ sed -i \'s/at 0x[0-9a-f]*//g\' freeze.*.log\n    # This removes pointer values that appear in the `repr(...)`.\n\n    $ meld freeze.0.log freeze.1.log\n    # Alternatively, use `icdiff` or `diff -u1`.\n\nTODO\n----\n\n- ☑ Bring hash into separate package.\n\n- ☐ Correctness\n\n  - ☑ Test hashing sets with different orders. Assert tests fail.\n  - ☑ Test hashing dicts with different orders. Assert tests fail.\n  - ☑ Don\'t include properties in hash.\n  - ☐ Support closures which include `import x` and `from x import y`\n  - ☐ Test for unique representation between types.\n  - ☐ Test functions with minor changes.\n  - ☐ Test set/dict with diff hash.\n  - ☐ Test obj with slots.\n  - ☐ Test hash for objects and classes more carefully.\n  - ☑ Improve test coverage.\n  - ☐ Investigate when modules are assumed constant.\n\n- ☑ API\n\n  - ☑ Use user-customizable multidispatch.\n  - ☑ Make it easier to register a freeze method for a type.\n  - ☑ Encapsulate global config `freeze` into object.\n  - ☐ Move "get call graph" into its own package.\n\n- ☐ Make ``freeze`` handle more types:\n\n  - ☑ Module: freeze by name.\n  - ☑ Objects: include the source-code of methods.\n  - ☑ C extensions. freeze by name, like module\n  - ☑ Methods\n  - ☑ fastpath for numpy arrays\n  - ☑ ``tqdm``\n  - ☑ ``numpy.int64(1234)``\n  - ☑ Pandas dataframe\n  - ☑ Catch Pickle TypeError\n  - ☑ Catch Pickle ImportError\n\n- ☐ Performance\n\n  - ☐ Make performance benchmarks.\n  - ☐ Memoize the hash of immutable data.\n\n    - Note: this might not work, because immutable data can still contain pointers to mutable data, e.g.: in ``a = ([])``, ``a`` is immutable, but ``a[0]`` is not.\n',
    'author': 'Samuel Grayson',
    'author_email': 'sam+dev@samgrayson.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/charmoniumQ/charmonium.freeze',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
